#!/usr/bin/ruby

require 'fcgi'
require 'logger'
require 'optparse'
require 'ostruct'
require 'yaml'
require 'digest/md5'
require 'open-uri'
require 'tempfile'
require 'timeout'
require 'ftools'
require 'uri'

def get_lock(log, file, timeout)
  lock_name = file + '.lock'
  lock_file = nil
  # create file if it does not exist yet
  open(lock_name, 'a', 0600) if not File.exists?(lock_name)
  log.debug("Attempting to get lock ...")
  begin
    Timeout::timeout(timeout + timeout / 10) {
      while not (lock_file = open(lock_name, 'r', 0600)) or not lock_file.flock(File::LOCK_EX | File::LOCK_NB)
        delay = timeout / 10
        log.debug("Failed go get lock, waiting #{delay} seconds ...")
        sleep(delay)
      end
    }
  rescue TimeoutError
    log.error("Failed to get lock, had some other process hung?")
    raise "Failed to get lock"
  end
  return lock_file
end

def retrieve_file(log, tmp_file, req_url, timeout)

  log.debug("Requesting file via url #{req_url} into temp file #{tmp_file.path}")
  meta = nil

  begin
    Timeout::timeout(timeout) {
      open(req_url) { |f|
        meta = f.meta
        log.debug("Meta info: #{meta.inspect}")
        tmp_file.write f.read
      }
    }
  rescue Timeout::Error
    log.debug("timeout while fetching file")
    return nil
  rescue
    log.debug("generic error while fetching the file: #{$!}")
    return nil
  end

  return meta

end

version = "0.1_alpha"
options = OpenStruct.new
options.config = '/etc/gravacacher/gravacacher.yaml'
options.log = '/var/log/gravacacher/gravacacher.log'
options.log_level = Logger::WARN

LOG_LEVELS = {
              'DEBUG' => Logger::DEBUG,
              'INFO' => Logger::INFO,
              'WARN' => Logger::WARN,
              'ERROR' => Logger::ERROR,
              'FATAL' => Logger::FATAL
             }

opts = OptionParser.new do |opts|
  opts.banner = "Usage: gravacacher.rb [options]"

  opts.separator ""
  opts.separator "Specific options:"

  opts.on("-c", "--config CONFIGFILE", "Config file to use") do |cfg|
    options.config = cfg
  end

  opts.on("-l", "--log LOGFILE", "Log file to use") do |log|
    options.log = log
  end

  opts.on("--log-level [LEVEL]", LOG_LEVELS.keys,
          "Select logger level (DEBUG, INFO, WARN, ERROR, FATAL)") do |log_level|
    options.log_level = LOG_LEVELS[log_level]
  end

  opts.separator ""
  opts.separator "Common options:"

  opts.on_tail("-v", "--version", "Print software version") do
    puts "gravacacher #{version}"
    exit
  end

  opts.on_tail("-h", "--help", "Show this message") do
    puts opts
    exit
  end
 
  opts.parse!(ARGV)
  options
end

config = YAML::load File.new(options.config)
log = Logger.new(options.log, 10, 1024*1024)
log.level = options.log_level

log.debug("Starting processing")

password = ''
password = config['general']['password'] if config['general'].has_key?('password')

if password == ''
  log.warn("Password is not specified in config - this means you're in fact running open proxy server!")
end

FCGI.each {|request|
  out = request.out

  catch(:finish) do
    begin
      gravacacher_id = ''
      server = ''
      server_hash = ''
      params = ''
      default_url = ''
      query_opts = Hash.new

      query = request.env['QUERY_STRING'].dump.sub(/^\"(.*)\"$/, '\1')
      log.debug("Parsing query #{query}")
      query.split('&').each { |pair|
        (key, val) = pair.split('=')
	log.debug("key=#{key.dump}, val=#{val.dump}")
        if key =~ /[^0-9a-zA-Z_%\.]/ or val =~ /[^0-9a-zA-Z_%\.]/
	  log.error("Invalid key or value: #{key}=#{val} in query #{query}")
	  out.print "Content-type: text/plain\r\n"
          out.print "Status: 400 Bad Request\r\n\r\n"
	  out.print "Fatal error: incorrect characters in parameters.\r\n"
	  throw :finish
        end
	query_opts[key] = val
      }

      # Pass 1: find server key
      query_opts.each { |key, val|
        if config['servers'].has_key?(key)
	  log.debug("Using server #{key} with hash #{val}")
	  server = key
	  server_hash = val
	  break
	end
      }

      if server == ''
        log.error("'server' parameter was not found in query #{query}")
	out.print "Content-type: text/plain\r\n"
	out.print "Status: 400 Bad Request\r\n\r\n"
	out.print "Fatal error: wrong parameters or their order.\r\n"
	throw :finish
      end

      # Pass 2: get all other parameters
      query_opts.each { |key, val|
	if key == 'gravacacher_id'
	  log.debug("Cacher requested hash is #{val}")
	  gravacacher_id = val
	elsif key == server
	  # Nothing to do, parsed already
	else
	  allowed_params = config['servers'][server]['params'].split(' ')
	  if not allowed_params.find { |p| key == p }
	    log.error("Unknown key #{key} in query #{query}")
	    out.print "Content-type: text/plain\r\n"
	    out.print "Status: 400 Bad Request\r\n\r\n"
	    out.print "Fatal error: unknown key #{key}.\r\n"
	    throw :finish
	  end
	  log.debug("Adding parameters #{key}=#{val}")
	  params += "&#{key}=#{val}"
	  if key == 'default'
	    default_url = val
	  end
	end
      }

      if (gravacacher_id == '' and password != '') or server == '' or server_hash == ''
        log.error("Missing one of required parameters, gravacacher_id=#{gravacacher_id}, server=#{server}, server_hash=#{server_hash}, params=#{params}")
        out.print "Content-type: text/plain\r\n"
        out.print "Status: 400 Bad Request\r\n\r\n"
        out.print "Fatal error: missing one of required parameters.\r\n"
        throw :finish
      end

      log.debug("Processing: gravacacher_id=#{gravacacher_id}, server=#{server}, server_hash=#{server_hash}, params=#{params}")

      # Is it possible to do it nicer somehow?
      in_str = query.sub(/(^gravacacher_id=[0-9a-zA-Z]+&)/, '')
      in_str = in_str.sub(/(&gravacacher_id=[0-9a-zA-Z]+)/, '')

      log.debug("Calculating digest of #{in_str}")
      our_id = Digest::MD5.hexdigest(password + in_str)
      log.debug("Digest is: #{our_id}")

      if password == ''
        log.debug("As password is not given - reset gravacacher_id to our id")
	gravacacher_id = our_id
      end

      if our_id != gravacacher_id
        log.error("Wrong gravacacher_id in query #{query}")
        out.print "Content-type: text/plain\r\n"
        out.print "Status: 403 Forbidden\r\n\r\n"
        out.print "Fatal error: wrong gravacacher_id specified\n"
        throw :finish
      end

      permanent_file = config['general']['permanent_dir'] + '/' + gravacacher_id
      cached_file = config['general']['cached_dir'] + '/' + gravacacher_id
      file_path = cached_file
      fl_retrieve = true
      fl_lock_needed = true
      timeout = config['servers'][server]['retrieval_timeout']
      lock_file = nil
      neg_file_name = cached_file + '.neg'

      begin # retrieval + answer block
        if File::exists?(permanent_file)
          log.info("Found file in permanent storage")
	  fl_retrieve = false
	  fl_lock_needed = false
	  file_path = permanent_file
	else
	  lock_file = get_lock(log, cached_file, timeout)
          if File::exists?(cached_file)
	    timeout = config['servers'][server]['update_timeout']
            mtime = File.mtime(cached_file)
	    curtime = Time.new
	    log.debug("File modification time is #{mtime}, current time is #{curtime}")
	    if mtime + config['servers'][server]['ttl'] > curtime
	      log.debug("TTL has not expired yet, no update is necessary")
	      fl_retrieve = false
	    end
          end
	end

	if fl_retrieve
	  if File::exists?(neg_file_name)
	    mtime = File.mtime(neg_file_name)
	    curtime = Time.new
	    log.debug("Negative response flag found, dated #{curtime}")
	    if mtime + config['servers'][server]['negttl'] <= curtime
	      log.debug("NegTTL has expired, removing flag")
	      File.unlink(neg_name)
	    else
	      fl_retrieve = false
	    end
	  end
	end

        if fl_retrieve
          log.debug("No suitable file found in permanent or cache storages: downloading")
          tmp_file = nil

          begin # retrieval block
            tmp_dir = config['general']['cached_dir']
            log.debug("Creating temp file in #{tmp_dir} with basename #{gravacacher_id}")
            tmp_file = Tempfile.new(gravacacher_id, tmp_dir)
            req_url = "#{config['servers'][server]['url']}?#{server}=#{server_hash}#{params}"

	    meta = retrieve_file(log, tmp_file, req_url, timeout)

	    log.debug("Main retrieval failed, trying default") if not meta
	    meta = retrieve_file(log, tmp_file, URI.unescape(default_url), timeout) if not meta and default_url != ''

	    raise "failed to retrieve file" if not meta

	    open(cached_file + '.meta', 'wb') { |file|
              if meta.has_key?('content-type')
                file.write "Content-Type: #{meta['content-type']}\r\n"
              end
              if meta.has_key?('last-modified')
                file.write "Last-Modified: #{meta['last-modified']}\r\n"
              end
	    }

	    File.rename(tmp_file.path, cached_file)
	    tmp_file.close
          rescue # retrieval block
            log.warn("Cannot download file into cache: #{$!}, query #{query}")
	    tmp_file.unlink
	    log.debug("Creating .neg flag file")
	    open(neg_file_name, 'a', 0600)
          end # retrieval block
        end # if fl_retrieve

        if not File::exists?(file_path)
          log.warn("Was not able to serve query #{query}")
	  out.print "Status: 504 Gateway Timeout\r\n\r\n"
	  throw :finish
        end

        if File::exists?(file_path + '.meta')
          open(file_path + '.meta', 'rb') { |file|
            out.write file.read
	  }
        end

        file_size = File.size(file_path)
        log.debug("Generating output, file path is #{file_path}, size is #{file_size}")
        out.print "Content-Length: #{file_size}\r\n\r\n"

        open(file_path, 'rb') { |file|
          out.write file.read
        }

        log.debug("Finished serving request succesfully")
      rescue
        log.error("Fatal: answer block internal error: #{$!} in query #{query}")
      ensure # retrieve + answer block
        if lock_file
          log.debug("Releasing lock file")
	  lock_file.flock(File::LOCK_UN)
	  lock_file.close
	end
      end # retrieve + answer block
    rescue # the whole request block
      log.error("Fatal: internal error: #{$!} in query #{query}")
      out.print "Status: 500 Internal server error\r\n\r\n"
      throw :finish
    end # the whole request block
  end # catch(:finish) block

  request.finish
}
